"""Meta-estimators for composing models with multiple transformers.

These estimators are useful for working with heterogenous tabular data.
"""

from ._column_transformer import ColumnTransformer, make_column_transformer

__all__ = ["ColumnTransformer", "make_column_transformer"]



print("import: 'dask_ml'")
import dask_ml

print("import: 'dask_ml.cluster'")
import dask_ml.cluster

print("import: 'dask_ml.decomposition'")
import dask_ml.decomposition

print("import: 'dask_ml.feature_extraction'")
import dask_ml.feature_extraction

print("import: 'dask_ml.linear_model'")
import dask_ml.linear_model

print("import: 'dask_ml.metrics'")
import dask_ml.metrics

print("import: 'dask_ml.model_selection'")
import dask_ml.model_selection

print("import: 'dask_ml.preprocessing'")
import dask_ml.preprocessing

print("import: 'dask_ml.utils'")
import dask_ml.utils

print("import: 'dask_ml.wrappers'")
import dask_ml.wrappers




from typing import Any, Callable, Union

import dask.array as da
import dask.dataframe as dd
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin

from dask_ml.utils import check_array

from .._typing import ArrayLike, DataFrameType, SeriesType


class BlockTransformer(BaseEstimator, TransformerMixin):
    """Construct a transformer from a an arbitrary callable

    The BlockTransformer forwards the blocks of the X arguments to a user-defined
    callable and returns the result of this operation.
    This is useful for stateless operations, that can be performed on the cell or
    block level, such as taking the log of frequencies. In general the transformer
    is not suitable for e.g. standardization tasks as this requires information for
    a complete column.

    Parameters
    ----------
    func : callable
        The callable to use for the transformation.

    validate : bool, optional default=False
         Indicate that the input X array should be checked before calling
        ``func``.

    kw_args : dict, optional
        Dictionary of additional keyword arguments to pass to func.

    Examples
    --------

    >>> import dask.datasets
    >>> import pandas as pd
    >>> from dask_ml.preprocessing import BlockTransformer
    >>> df = dask.datasets.timeseries()
    >>> df
    ... # doctest: +SKIP
    Dask DataFrame Structure:
                       id    name        x        y
    npartitions=30
    2000-01-01      int64  object  float64  float64
    2000-01-02        ...     ...      ...      ...
    ...               ...     ...      ...      ...
    2000-01-30        ...     ...      ...      ...
    2000-01-31        ...     ...      ...      ...
    Dask Name: make-timeseries, 30 tasks
    >>> trn = BlockTransformer(pd.util.hash_pandas_object, index=False)
    >>> trn.transform(df)
    ... # doctest: +ELLIPSIS
    Dask Series Structure:
    npartitions=30
    2000-01-01    uint64
    2000-01-02       ...
                ...
    2000-01-30       ...
    2000-01-31       ...
    dtype: uint64
    Dask Name: hash_pandas_object, 60 tasks
    """

    def __init__(
        self,
        func: Callable[..., Union[ArrayLike, DataFrameType]],
        *,
        validate: bool = False,
        **kw_args: Any,
    ):
        self.func: Callable[..., Union[ArrayLike, DataFrameType]] = func
        self.validate = validate
        self.kw_args = kw_args

    def fit(
        self, X: Union[ArrayLike, DataFrameType], y: Union[ArrayLike, SeriesType] = None
    ) -> "BlockTransformer":
        return self

    def transform(
        self, X: Union[ArrayLike, DataFrameType], y: Union[ArrayLike, SeriesType] = None
    ) -> Union[ArrayLike, DataFrameType]:
        kwargs = self.kw_args if self.kw_args else {}

        if isinstance(X, da.Array):
            if self.validate:
                X = check_array(X, accept_dask_array=True, accept_unknown_chunks=True)
            XP = X.map_blocks(self.func, dtype=X.dtype, chunks=X.chunks, **kwargs)
        elif isinstance(X, dd.DataFrame):
            if self.validate:
                X = check_array(
                    X, accept_dask_dataframe=True, preserve_pandas_dataframe=True
                )
            XP = X.map_partitions(self.func, **kwargs)
        elif isinstance(X, pd.DataFrame):
            if self.validate:
                X = check_array(
                    X, accept_dask_array=False, preserve_pandas_dataframe=True
                )
            XP = self.func(X, **kwargs)
        else:
            if self.validate:
                X = check_array(X, accept_dask_array=False)
            XP = self.func(X, **kwargs)
        return XP



from .classification import accuracy_score, log_loss  # noqa
from .pairwise import (  # noqa
    euclidean_distances,
    pairwise_distances,
    pairwise_distances_argmin_min,
)
from .regression import (  # noqa
    mean_absolute_error,
    mean_absolute_percentage_error,
    mean_squared_error,
    mean_squared_log_error,
    r2_score,
)
from .scorer import SCORERS, check_scoring, get_scorer  # noqa



from __future__ import absolute_import, division, print_function

import warnings

import numpy as np
import sklearn.exceptions
from dask.base import normalize_token
from sklearn.base import BaseEstimator
from sklearn.model_selection._split import (
    BaseShuffleSplit,
    LeaveOneGroupOut,
    LeaveOneOut,
    LeavePGroupsOut,
    LeavePOut,
    PredefinedSplit,
    _BaseKFold,
    _CVIterableWrapper,
)


@normalize_token.register(BaseEstimator)
def normalize_estimator(est):
    """Normalize an estimator.

    Note: Since scikit-learn requires duck-typing, but not sub-typing from
    ``BaseEstimator``, we sometimes need to call this function directly."""
    base = [type(est).__name__, normalize_token(est.get_params())]
    # fitted attributes: https://github.com/dask/dask-ml/issues/658
    attrs = [x for x in dir(est) if x.endswith("_") and not x.startswith("_")]
    exclude = {"cv_results_", "model_history_", "history_", "refit_time_"}

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", FutureWarning)
        for attr in attrs:
            if attr in exclude:
                continue
            try:
                val = getattr(est, attr)
            except (sklearn.exceptions.NotFittedError, AttributeError):
                continue
            base.append(val)
    return tuple(base)


def normalize_random_state(random_state):
    if isinstance(random_state, np.random.RandomState):
        return random_state.get_state()
    return random_state


@normalize_token.register(_BaseKFold)
def normalize_KFold(x):
    # Doesn't matter if shuffle is False
    rs = normalize_random_state(x.random_state) if x.shuffle else None
    return (type(x).__name__, x.n_splits, x.shuffle, rs)


@normalize_token.register(BaseShuffleSplit)
def normalize_ShuffleSplit(x):
    return (
        type(x).__name__,
        x.n_splits,
        x.test_size,
        x.train_size,
        normalize_random_state(x.random_state),
    )


@normalize_token.register((LeaveOneOut, LeaveOneGroupOut))
def normalize_LeaveOneOut(x):
    return type(x).__name__


@normalize_token.register((LeavePOut, LeavePGroupsOut))
def normalize_LeavePOut(x):
    return (type(x).__name__, x.p if hasattr(x, "p") else x.n_groups)


@normalize_token.register(PredefinedSplit)
def normalize_PredefinedSplit(x):
    return (type(x).__name__, x.test_fold)


@normalize_token.register(_CVIterableWrapper)
def normalize_CVIterableWrapper(x):
    return (type(x).__name__, x.cv)
